from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from zope.app.component import hooks
from zope.i18n import translate


class BaseSubjectCreator(object):
    """An Overwritable Adapter to get the Subject"""

    def __init__(self, context):

        self.context = aq_inner(context)
        self.request = self.context.REQUEST

    def __call__(self, object_):
        site = hooks.getSite()
        site_title = site.Title()
        if isinstance(site_title, unicode):
            site_title = site_title.encode('utf-8')
        portal_properties = getToolByName(object_, 'portal_properties')
        default_subject = '[%s] %s: %s' % (
            site_title,
            translate(u'Notification',
                     domain="ftw.notification.email",
                     context=object_.REQUEST).encode('utf8'),
            object_.Title())
        subject = None
        try:
            sheet = portal_properties.ftw_notification_properties
        except AttributeError:
            subject = default_subject
        else:
            subject = sheet.getProperty(
                'notification_email_subject',
                default_subject) + ': ' + object_.Title()
        return subject


class AttachmentCreator(object):
    """An base-adapter to add attachments"""

    def __init__(self, context):
        self.context = aq_inner(context)
        self.request = self.context.REQUEST

    def __call__(self, object_):
        """Usualy this Adapters does nothing,
        bewlow is an example how implement an attachment using StringIO.

        @return tuple(file, name, mimetype)"""

        # mime_type = "text/calendar"
        # attachments = []
        # self.meeting_types = ['dates_additional',
        #                       'meeting_dates_additional',]

        # if object_.getMeeting_type() in self.meeting_types:
        #     ical_view = getattr(object_.aq_explicit, 'ics_view', None)
        #     if ical_view:
        #         ical_view = ical_view(
        #             object_.REQUEST, object_.REQUEST.RESPONSE)
        #         ical_file = StringIO(ical_view)
        #         ical_attachment = (
        #             ical_file, 'ICal.ics', mime_type.split('/'))
        #         attachments = [ical_attachment,]
        # return attachments
        return []



from ftw.notification.base import notification_base_factory as _nb
from ftw.notification.base.notifier import BaseNotifier
from ftw.notification.email.interfaces import (
    IEMailRepresentation, ISubjectCreator, IAttachmentCreator)
from Products.CMFCore.utils import getToolByName
from Products.statusmessages.interfaces import IStatusMessage
from StringIO import StringIO
from ZODB.POSException import ConflictError
from zope.app.component import hooks
from zope.publisher.interfaces import Retry
import logging
import sys
import traceback


logger = logging.getLogger('ftw.notification.email')

_marker = object()


class MailNotifier(BaseNotifier):

    def create_recipients(self, user_list):
        """Creates a unique list of recipients"""

        site = hooks.getSite()
        portal_membership = getToolByName(site, 'portal_membership')

        recipients = {}

        for user_id in user_list:
            member = portal_membership.getMemberById(user_id)
            if member is None:
                continue
            fullname = member.getProperty('fullname', user_id)
            if not len(fullname):
                fullname = user_id
            email = member.getProperty('email', None)
            if email is None:
                continue
            recipients[user_id] = (fullname, email)
        return recipients

    def send_notification(self, to_list=_marker, cc_list=_marker,
                          object_=None, message=u"", **kwargs):

        if to_list is _marker:
            to_list = []

        if cc_list is _marker:
            cc_list = []

        site = hooks.getSite()
        portal_membership = getToolByName(object_ or site, 'portal_membership')

        recipients = self.create_recipients(to_list)
        cc_recipients = self.create_recipients(cc_list)

        if not recipients:
            return

        sender = None
        sender_id = kwargs.get('actor', '')
        sender_data = portal_membership.getMemberById(sender_id)
        if sender_data is not None:
            sender_fullname = sender_data.getProperty('fullname', sender_id)
            if not len(sender_fullname):
                sender_fullname = sender_id
            sender_email = sender_data.getProperty('email', '')
            sender = (sender_fullname, sender_email)
        kwargs.update(dict(sender=sender))
        if object_ is not None:
            try:
                subject = ISubjectCreator(object_)(object_)
                # Call attachment adapter
                attachments = IAttachmentCreator(object_)(object_)
                # subject should be utf-8
                if isinstance(subject, unicode):
                    subject = subject.encode('utf-8')
                email = IEMailRepresentation(object_)(subject,
                                                      recipients.values(),
                                                      cc_recipients.values(),
                                                      message,
                                                      attachments=attachments,
                                                      **kwargs)
                mailhost = getToolByName(object_, "MailHost")

                # XXX: Unfortunality we have to implement the carbon copy
                # feature by ourself.
                to_addr = '%s, %s' % (email['To'], email['CC'])
                mailhost.send(
                    email.as_string(), to_addr, email['From'], subject)
                IStatusMessage(object_.REQUEST).addStatusMessage(
                        _nb('statusmessage_notification_sent'), type='info')
            except (ConflictError, Retry):
                raise
            except Exception, error:
                exceptionType, exceptionValue, exceptionTraceback = \
                                                            sys.exc_info()
                exs = StringIO()
                print error
                exs.write('Error while sending notification\n')
                traceback.print_exception(exceptionType, exceptionValue,
                                          exceptionTraceback, file=exs)
                exs.seek(0)
                logger.error(exs.read())
                IStatusMessage(object_.REQUEST).addStatusMessage(
                    _nb('statusmessage_notification_not_sent'), type='error')



from Acquisition import aq_inner
from ftw.notification.email.composer import HTMLComposer
from zope.app.pagetemplate import ViewPageTemplateFile


_marker = object()


class BaseEmailRepresentation(object):
    """returns a email representation of an IBaseContent for HTML.
    """

    template = ViewPageTemplateFile('base.pt')
    #XXX: replace with CSSVIew
    stylesheet = """
    body {
        font-family: Arial !important;
        font-size: 10pt !important;
    }
    h1, h2, h3, h4, h5, h6 {
        border: none;
        font-family:  Arial !important;
    }
    div, p, ul, dl, ol {
        width: auto;
    }
    ul, ol, dl {
        padding-right: 0.5em;
    }
    ul {
        list-style-type: square;
    }
    .documentDescription {
        font-weight: bold;
    }
    pre {
        border: 1pt dotted black;
        white-space: pre;
        overflow: auto;
        padding: 1em 0;
    }
    table.listing,
    table.listing td {
        border: 1pt solid black;
        border-collapse: collapse;
    }
    a {
        color: Black !important;
        padding: 0 !important;
        text-decoration: underline !important;
    }
    a:link, a:visited {
        color: #520;
        background: transparent;
    }

    hr{
        border:none;
        border-bottom:1px solid grey;
    }
    """

    def __init__(self, context):
        self.context = aq_inner(context)
        self.request = self.context.REQUEST
        self.comment = None

    def __call__(self, subject, to_list, cc_list, message,
                 attachments=_marker, **kwargs):

        if attachments is _marker:
            attachments = []

        self.comment = message.replace('\n', '<br />')
        self.__dict__.update(kwargs)
        composer = HTMLComposer(self.template(self), subject, to_list, cc_list,
                                replyto_address=kwargs['sender'])
        return composer.render(filter_tags=False,
                               override_vars=dict(stylesheet=self.stylesheet),
                               attachments=attachments)



from DateTime import DateTime
from ftw.journal.events.events import JournalEntryEvent
from ftw.notification.base import notification_base_factory as _
from ftw.notification.base.interfaces import INotifier
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.event import notify
from zope.i18n import translate


def notification_sent(event):
    obj = event.obj
    comment = event.comment
    notifier = queryUtility(INotifier, name='email-notifier')

    if notifier is None:
        return

    if event.action is None:
        action = _(u"label_send_notification", default=u"Send Notification")
    else:
        action = event.action

    if event.actor is None:
        portal_state = getMultiAdapter((obj, obj.REQUEST),
                                        name=u'plone_portal_state')
        actor = portal_state.member().getId()
    else:
        actor = event.actor

    if event.time is None:
        time = DateTime()
    else:
        time = event.time

    to_list = obj.REQUEST.get('to_list', [])
    cc_list = obj.REQUEST.get('cc_list', [])

    journal_comment = translate(
        msgid=u'journal_notification_text',
        domain='ftw.notification.email',
        context=obj.REQUEST,
        mapping=dict(
            to_list=len(to_list) > 0 and ', '.join(to_list) + '\n' or '-',
            cc_list=len(cc_list) > 0 and ', '.join(cc_list) + '\n' or '-',
            comment=comment))

    notify(JournalEntryEvent(obj, journal_comment, action))

    kwargs = dict(action=action, actor=actor, time=time)
    notifier.send_notification(
        to_list=to_list,
        cc_list=cc_list,
        message=comment,
        object_=obj,
        **kwargs)



from Products.PloneTestCase import ptc
from ftw.notification.email.testing import ftw_notification_layer


class FtwNotificationTestCase(ptc.PloneTestCase):
    """Base class for integration tests."""

    layer = ftw_notification_layer



import doctest
from plone.testing import layered
import unittest2 as unittest
from ftw.notification.email.testing import NOTIFICATION_INTEGRATION_TESTING


OPTIONFLAGS = (doctest.NORMALIZE_WHITESPACE|
               doctest.ELLIPSIS|
               doctest.REPORT_NDIFF)


TESTFILES = (
    'notification.txt',
    )



def test_suite():
    suite = unittest.TestSuite()
    suite.addTests([layered(doctest.DocFileSuite(filename,
                                                 optionflags=OPTIONFLAGS),
                    layer=NOTIFICATION_INTEGRATION_TESTING)
                    for filename in TESTFILES])
    return suite


from plone.app.testing import applyProfile
from plone.app.testing import IntegrationTesting, FunctionalTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from zope.configuration import xmlconfig


class NotificationIntegrationLayer(PloneSandboxLayer):
    """Layer for integration tests."""

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):

        # Load testing zcml (optional)
        import ftw.notification.base
        xmlconfig.file('configure.zcml', ftw.notification.base,
                       context=configurationContext)

        import ftw.notification.email
        xmlconfig.file('configure.zcml', ftw.notification.email,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.notification.email:default')


NOTIFICATION_INTEGRATION_FIXTURE = NotificationIntegrationLayer()
NOTIFICATION_INTEGRATION_TESTING = IntegrationTesting(
    bases=(NOTIFICATION_INTEGRATION_FIXTURE, ),
    name="Notification:Integration")

NOTIFICATION_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(NOTIFICATION_INTEGRATION_FIXTURE,),
    name="Notification:Functional")



from email import Encoders
from email import Utils
from email.Header import Header
from email.MIMEBase import MIMEBase
from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from email.Utils import formatdate
from ftw.notification.email.interfaces import IEMailComposer
from zope import component
from zope import interface
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
import Products.CMFPlone.interfaces
import StringIO
import formatter
import htmllib
import persistent
import stoneagehtml
import string
import zope.sendmail.mailer


_ = lambda x: x


def _render_cachekey(method, self, vars_):
    return (vars_)


_marker = object()


def create_html_mail(subject, html, text=None, from_addr=None, to_addr=None,
                     headers=None, encoding='UTF-8', attachments=_marker):
    """Create a mime-message that will render HTML in popular
    MUAs, text in better ones.
    """

    if attachments is _marker:
        attachments = []

    # Use DumbWriters word wrapping to ensure that no text line
    # is longer than plain_text_maxcols characters.
    plain_text_maxcols = 72
    html = html.encode(encoding)
    if text is None:
        # Produce an approximate textual rendering of the HTML string,
        # unless you have been given a better version as an argument
        textout = StringIO.StringIO()
        formtext = formatter.AbstractFormatter(formatter.DumbWriter(
                        textout, plain_text_maxcols))
        parser = htmllib.HTMLParser(formtext)
        parser.feed(html)
        parser.close()

        # append the anchorlist at the bottom of a message
        # to keep the message readable.
        counter = 0
        anchorlist = "\n\n" + ("-" * plain_text_maxcols) + "\n\n"
        for item in parser.anchorlist:
            counter += 1
            anchorlist += "[%d] %s\n" % (counter, item)

        text = textout.getvalue() + anchorlist
        del textout, formtext, parser, anchorlist
    else:
        text = text.encode(encoding)

    # if we would like to include images in future, there should
    # probably be 'related' instead of 'mixed'
    msg = MIMEMultipart('mixed')
    msg['Subject'] = Header(subject, encoding)
    msg['From'] = from_addr
    msg['To'] = to_addr
    msg['Date'] = formatdate(localtime=True)
    msg["Message-ID"] = Utils.make_msgid()
    if headers:
        for key, value in headers.items():
            msg[key] = value

    msg.preamble = 'This is a multi-part message in MIME format.'

    alternatives = MIMEMultipart('alternative')
    msg.attach(alternatives)
    alternatives.attach(MIMEText(text, 'plain', _charset=encoding))
    alternatives.attach(MIMEText(html, 'html', _charset=encoding))

    #add the attachments
    for f, name, mimetype in attachments:
        if mimetype is None:
            mimetype = ('application', 'octet-stream')
        maintype, subtype = mimetype
        if maintype == 'text':
            # XXX: encoding?
            part = MIMEText(f.read(), _subtype=subtype)
        else:
            part = MIMEBase(maintype, subtype)
            part.set_payload(f.read())
            Encoders.encode_base64(part)
        part.add_header(
            'Content-Disposition', 'attachment; filename="%s"' % name)
        msg.attach(part)
    return msg


class HTMLComposer(persistent.Persistent):
    # """
    #   >>> from zope.interface.verify import verifyClass
    #   >>> from collective.dancing.interfaces import IHTMLComposer
    #   >>> verifyClass(IHTMLComposer, HTMLComposer)
    #   True
    # """

    interface.implements(IEMailComposer)

    title = _(u'HTML E-Mail')

    def __init__(self, message, subject, to_addresses, cc_addresses=_marker,
        from_name=u'', from_address='', stylesheet='', replyto_address=''):

        if cc_addresses is _marker:
            cc_addresses = []

        properties = component.getUtility(
            Products.CMFCore.interfaces.IPropertiesTool)
        self.encoding = properties.site_properties.getProperty(
            'default_charset', 'utf-8')
        self.stylesheet = stylesheet
        self.from_name = from_name or properties.email_from_name
        self.from_address = from_address or properties.email_from_address
        self.to_addresses = to_addresses and to_addresses or None
        self.cc_addresses = cc_addresses
        self.replyto_address = replyto_address
        self.subject = subject
        self.message = message
        self.header_text = u""
        self.footer_text = u""

    template = ViewPageTemplateFile('templates/composer-html.pt')

    context = None

    @property
    def request(self):
        site = zope.app.component.hooks.getSite()
        return site.REQUEST

    def _prepare_address(self, name, mail, charset):
        if not isinstance(name, unicode):
            name = name.decode(charset)
        if not isinstance(mail, unicode):
            # mail has to be be ASCII!!
            mail = mail.decode(charset).encode('us-ascii', 'replace')
            #TODO : assert that mail is now valid. (could have '?' from repl.)
        return Utils.formataddr((str(Header(name, charset)), mail))

    @property
    def _from_address(self):
        return self._prepare_address(
            self.from_name, self.from_address, self.encoding)

    def _to_addresses(self, to_addresses):
        addresses = []
        for name, mail in to_addresses:
            addresses.append(self._prepare_address(name, mail, self.encoding))
        return ', '.join(addresses)

    @property
    def _replyto_address(self):
        name, mail = self.replyto_address
        return self._prepare_address(name, mail, self.encoding)

    @property
    def language(self):
        return self.request.get('LANGUAGE')

    def _vars(self):
        """Provide variables for the template.

        Override this or '_more_vars' in your custom HTMLComposer to
        pass different variables to the templates.
        """
        variables = {}
        site = component.getUtility(
            Products.CMFPlone.interfaces.IPloneSiteRoot)
        #site = utils.fix_request(site, 0)
        #lambda t: transform.URL(site).__call__(t, subscription)
        fix_urls = lambda t: t
        variables['site_url'] = site.absolute_url()
        variables['site_title'] = site.Title()
        variables['subject'] = self.subject
        variables['message'] = self.message
        # Why would header_text or footer_text ever be None?
        variables['header_text'] = fix_urls(self.header_text or u"")
        variables['footer_text'] = fix_urls(self.footer_text or u"")
        variables['stylesheet'] = self.stylesheet
        variables['from_addr'] = self._from_address
        variables['to_addr'] = self._to_addresses(self.to_addresses)
        headers = variables['more_headers'] = {}
        if self.replyto_address:
            headers['Reply-To'] = self._replyto_address
        cc_addr = self._to_addresses(self.cc_addresses)
        if cc_addr:
            headers['CC'] = cc_addr

        # It'd be nice if we could use an adapter here to override
        # variables.  We'd probably want to pass 'items' along to that
        # adapter.

        return variables

    #@volatile.cache(_render_cachekey)
    def _html(self, variables, **kwargs):
        html = self.template(**variables)
        return stoneagehtml.compactify(html, **kwargs).decode('utf-8')

    def html(self, override_vars=None, template_vars=_marker, **kwargs):
        if template_vars is _marker:
            template_vars = {}

        variables = self._vars()

        if override_vars is None:
            override_vars = {}
        variables.update(override_vars)

        html = self._html(variables, **kwargs)
        html = string.Template(html).safe_substitute(template_vars)
        return html

    def render(self, override_vars=None, template_vars=_marker,
               attachments=_marker, **kwargs):

        if template_vars is _marker:
            template_vars = {}

        if attachments is _marker:
            attachments = []

        variables = self._vars()

        if override_vars is None:
            override_vars = {}
        variables.update(override_vars)
        message = create_html_mail(
            variables['subject'],
            self.html(
                override_vars=override_vars,
                template_vars=template_vars,
                **kwargs),
            from_addr=variables['from_addr'],
            to_addr=variables['to_addr'],
            headers=variables.get('more_headers'),
            encoding=self.encoding,
            attachments=attachments)

        return message



# pylint: disable=E0211, E0213
# E0211: Method has no argument
# E0213: Method should have "self" as first argument

from zope.interface import Interface
from ftw.notification.base.interfaces import INotifier


class IFormatItem(Interface):
    """A view that formats an item for use in a notification.

    Given an item as retrieved from a collector, this view returns a
    representation of the given item ready for inclusion in the
    message via the ``IComposer.render`` method.
    """

    def __call__():
        """Returns a unicode-string."""


class IEMailComposer(Interface):
    """An email composer.
    """


class ISubjectCreator(Interface):
    """Interface for
    """


class IMailNotifier(INotifier):
    """Interface for
    """


class IAttachmentCreator(Interface):
    """Interface for attachments
    """


class IEMailRepresentation(Interface):
    """Interface for
    """



from zope.i18nmessageid import MessageFactory

emailNotificationMessageFactory = MessageFactory('ftw.notification.email')


def initialize(context):
    """Initializer called when used as a Zope 2 product."""
